

public class Parser {
	public static final int _EOF = 0;
	public static final int _ident = 1;
	public static final int _number = 2;
	public static final int maxT = 28;

	static final boolean _T = true;
	static final boolean _x = false;
	static final int minErrDist = 2;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;
	
	public Scanner scanner;
	public Errors errors;

	Proc curProc;  // current program unit (procedure or main program)



	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (String msg) {
		if (errDist >= minErrDist) errors.SemErr(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) {
				++errDist;
				break;
			}

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	boolean StartOf (int s) {
		return set[s][la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}
	
	boolean WeakSeparator (int n, int syFol, int repFol) {
		int kind = la.kind;
		if (kind == n) { Get(); return true; }
		else if (StartOf(repFol)) return false;
		else {
			SynErr(n);
			while (!(set[syFol][kind] || set[repFol][kind] || set[0][kind])) {
				Get();
				kind = la.kind;
			}
			return StartOf(syFol);
		}
	}
	
	void Taste() {
		String name; 
		Expect(3);
		name = Ident();
		curProc = new Proc(name, null, this); 
		Expect(4);
		while (la.kind == 8 || la.kind == 9 || la.kind == 10) {
			if (la.kind == 8 || la.kind == 9) {
				VarDecl();
			} else {
				ProcDecl();
			}
		}
		Expect(5);
		for (Obj obj: curProc.locals) {
		if (obj instanceof Proc) ((Proc)obj).dump();
		} 
	}

	String  Ident() {
		String  name;
		Expect(1);
		name = t.val; 
		return name;
	}

	void VarDecl() {
		String name; Type type; 
		type = Typ();
		name = Ident();
		curProc.insert(new Var(name, type)); 
		while (la.kind == 6) {
			Get();
			name = Ident();
			curProc.insert(new Var(name, type)); 
		}
		Expect(7);
	}

	void ProcDecl() {
		String name; 
		Expect(10);
		name = Ident();
		Proc program = curProc;
		curProc = new Proc(name, program, this);
		program.insert(curProc); 
		Expect(11);
		Expect(12);
		curProc.block = Block();
		curProc = program; 
	}

	Type  Typ() {
		Type  type;
		type = Type.INT; 
		if (la.kind == 8) {
			Get();
		} else if (la.kind == 9) {
			Get();
			type = Type.BOOL; 
		} else SynErr(29);
		return type;
	}

	Block  Block() {
		Block  b;
		Stat s; 
		Expect(4);
		b = new Block(); 
		while (StartOf(1)) {
			if (StartOf(2)) {
				s = Stat();
				b.add(s); 
			} else {
				VarDecl();
			}
		}
		Expect(5);
		return b;
	}

	Stat  Stat() {
		Stat  s;
		String name; Expr e; Stat s2; Block b; 
		s = null; 
		switch (la.kind) {
		case 1: {
			name = Ident();
			Obj obj = curProc.find(name); 
			if (la.kind == 13) {
				Get();
				e = Expr();
				Expect(7);
				s = new Assignment((Var)obj, e); 
			} else if (la.kind == 11) {
				Get();
				Expect(12);
				Expect(7);
				s = new Call((Proc)obj); 
			} else SynErr(30);
			break;
		}
		case 14: {
			Get();
			Expect(11);
			e = Expr();
			Expect(12);
			s = Stat();
			s = new If(e, s); 
			if (la.kind == 15) {
				Get();
				s2 = Stat();
				s = new IfElse(s, s2); 
			}
			break;
		}
		case 16: {
			Get();
			Expect(11);
			e = Expr();
			Expect(12);
			s = Stat();
			s = new While(e, s); 
			break;
		}
		case 17: {
			Get();
			name = Ident();
			Expect(7);
			s = new Read((Var)curProc.find(name)); 
			break;
		}
		case 18: {
			Get();
			e = Expr();
			Expect(7);
			s = new Write(e); 
			break;
		}
		case 4: {
			b = Block();
			s = b; 
			break;
		}
		default: SynErr(31); break;
		}
		return s;
	}

	Expr  Expr() {
		Expr  e;
		Operator op; Expr e2; 
		e = SimExpr();
		if (la.kind == 25 || la.kind == 26 || la.kind == 27) {
			op = RelOp();
			e2 = SimExpr();
			e = new BinExpr(e, op, e2); 
		}
		return e;
	}

	Expr  SimExpr() {
		Expr  e;
		Operator op; Expr e2; 
		e = Term();
		while (la.kind == 19 || la.kind == 22) {
			op = AddOp();
			e2 = Term();
			e = new BinExpr(e, op, e2); 
		}
		return e;
	}

	Operator  RelOp() {
		Operator  op;
		op = Operator.EQU; 
		if (la.kind == 25) {
			Get();
		} else if (la.kind == 26) {
			Get();
			op = Operator.LSS; 
		} else if (la.kind == 27) {
			Get();
			op = Operator.GTR; 
		} else SynErr(32);
		return op;
	}

	Expr  Term() {
		Expr  e;
		Operator op; Expr e2; 
		e = Factor();
		while (la.kind == 23 || la.kind == 24) {
			op = MulOp();
			e2 = Factor();
			e = new BinExpr(e, op, e2); 
		}
		return e;
	}

	Operator  AddOp() {
		Operator  op;
		op = Operator.ADD; 
		if (la.kind == 22) {
			Get();
		} else if (la.kind == 19) {
			Get();
			op = Operator.SUB; 
		} else SynErr(33);
		return op;
	}

	Expr  Factor() {
		Expr  e;
		String name; 
		e = null; 
		if (la.kind == 1) {
			name = Ident();
			e = new Ident(curProc.find(name)); 
		} else if (la.kind == 2) {
			Get();
			e = new IntCon(Integer.parseInt(t.val)); 
		} else if (la.kind == 19) {
			Get();
			e = Factor();
			e = new UnaryExpr(Operator.SUB, e); 
		} else if (la.kind == 20) {
			Get();
			e = new BoolCon(true); 
		} else if (la.kind == 21) {
			Get();
			e = new BoolCon(false); 
		} else SynErr(34);
		return e;
	}

	Operator  MulOp() {
		Operator  op;
		op = Operator.MUL; 
		if (la.kind == 23) {
			Get();
		} else if (la.kind == 24) {
			Get();
			op = Operator.DIV; 
		} else SynErr(35);
		return op;
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		Taste();
		Expect(0);

		scanner.buffer.Close();
	}

	private static final boolean[][] set = {
		{_T,_x,_x,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x},
		{_x,_T,_x,_x, _T,_x,_x,_x, _T,_T,_x,_x, _x,_x,_T,_x, _T,_T,_T,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x},
		{_x,_T,_x,_x, _T,_x,_x,_x, _x,_x,_x,_x, _x,_x,_T,_x, _T,_T,_T,_x, _x,_x,_x,_x, _x,_x,_x,_x, _x,_x}

	};
} // end Parser


class Errors {
	public int count = 0;                                    // number of errors detected
	public java.io.PrintStream errorStream = System.out;     // error messages go to this stream
	public String errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
	
	protected void printMsg(int line, int column, String msg) {
		StringBuffer b = new StringBuffer(errMsgFormat);
		int pos = b.indexOf("{0}");
		if (pos >= 0) { b.delete(pos, pos+3); b.insert(pos, line); }
		pos = b.indexOf("{1}");
		if (pos >= 0) { b.delete(pos, pos+3); b.insert(pos, column); }
		pos = b.indexOf("{2}");
		if (pos >= 0) b.replace(pos, pos+3, msg);
		errorStream.println(b.toString());
	}
	
	public void SynErr (int line, int col, int n) {
		String s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "ident expected"; break;
			case 2: s = "number expected"; break;
			case 3: s = "\"program\" expected"; break;
			case 4: s = "\"{\" expected"; break;
			case 5: s = "\"}\" expected"; break;
			case 6: s = "\",\" expected"; break;
			case 7: s = "\";\" expected"; break;
			case 8: s = "\"int\" expected"; break;
			case 9: s = "\"bool\" expected"; break;
			case 10: s = "\"void\" expected"; break;
			case 11: s = "\"(\" expected"; break;
			case 12: s = "\")\" expected"; break;
			case 13: s = "\"=\" expected"; break;
			case 14: s = "\"if\" expected"; break;
			case 15: s = "\"else\" expected"; break;
			case 16: s = "\"while\" expected"; break;
			case 17: s = "\"read\" expected"; break;
			case 18: s = "\"write\" expected"; break;
			case 19: s = "\"-\" expected"; break;
			case 20: s = "\"true\" expected"; break;
			case 21: s = "\"false\" expected"; break;
			case 22: s = "\"+\" expected"; break;
			case 23: s = "\"*\" expected"; break;
			case 24: s = "\"/\" expected"; break;
			case 25: s = "\"==\" expected"; break;
			case 26: s = "\"<\" expected"; break;
			case 27: s = "\">\" expected"; break;
			case 28: s = "??? expected"; break;
			case 29: s = "invalid Typ"; break;
			case 30: s = "invalid Stat"; break;
			case 31: s = "invalid Stat"; break;
			case 32: s = "invalid RelOp"; break;
			case 33: s = "invalid AddOp"; break;
			case 34: s = "invalid Factor"; break;
			case 35: s = "invalid MulOp"; break;
			default: s = "error " + n; break;
		}
		printMsg(line, col, s);
		count++;
	}

	public void SemErr (int line, int col, String s) {	
		printMsg(line, col, s);
		count++;
	}
	
	public void SemErr (String s) {
		errorStream.println(s);
		count++;
	}
	
	public void Warning (int line, int col, String s) {	
		printMsg(line, col, s);
	}
	
	public void Warning (String s) {
		errorStream.println(s);
	}
} // Errors


class FatalError extends RuntimeException {
	public static final long serialVersionUID = 1L;
	public FatalError(String s) { super(s); }
}
